{-# Language TemplateHaskell,QuasiQuotes,FlexibleContexts,NoMonomorphismRestriction #-}

{-

This uses Peggy (http://tanakh.github.io/Peggy/ja/) .

Please do 

  cabal install Peggy

or

  cabal install Peggy --allow-newer

if your ghc is near the latest.

This is not an indent-oriented parser; to be replaced with better one.

-}

module Parser where
import Spec

import Text.Peggy hiding (Expr, space)


leftAssoc e oes = foldl (\e1 (o, e2) -> DFunAp (DBinOp o "") [e1,e2] "") e oes  

[peggy|
{- 
parseFile pRogramSpec "scc.txt"

parseString pRogramSpec "" "maxv g = let maxv_t v prev curr = prev v `max` maximum [ prev u | (e, u) <- is v ] in fregel val maxv_t Fix g" 

-}

pRogramSpecs :: [DProgramSpec String]
  = pRogramSpec pRogramSpec*                {  $1:$2  }

pRogramSpec :: DProgramSpec String   
  = rEcordSpec* pRog                            { DProgramSpec $1 $2 "" }
  
rEcordSpec :: DRecordSpec String
  = "data" cOnstructor "=" cOnstructor fIelds "deriving" "(" "Eq" "," "Show" ")" { DRecordSpec $2 $3 "" } 
 
fIelds :: [(DField String, DType String)]
  = "{" fieldSpec ("," fieldSpec)* "}"          { $1:$2 }

fieldSpec :: (DField String, DType String)
  = fIeld "::" tYpe                             { ($1, $2)}

pRog :: DProg String
  = fUn "g" "=" "let" gRoundDefs "in" gRaphExpr { DProg $1 $2 $3 "" }

gRoundDefs :: [DGroundDef String]
  = gRoundDef (";" gRoundDef)*                  { $1:$2 }

gRoundDef :: DGroundDef String
  = dEfVertInit                                 { DGDefVI $1 "" }
  / sMplDef                                     { DGDefSmpl $1 "" }
  / dEfGraphVar                                 { DGDefGV $1 "" }
  / dEfVertComp                                 { DGDefVC $1 "" }
  / dEfGraphFun                                 { DGDefGF $1 "" }
 
dEfVertComp :: DDefVertComp String
  = fUn "v" "prev" "curr" "=" "let" sMplDefs "in" eXpr { DDefVertComp $1 $2 $3 "" }
  / fUn "v" "prev" "curr" "=" eXpr              { DDefVertComp $1 [] $2 "" }

dEfVertInit :: DDefVertInit String
  = fUn "v" "=" "let" sMplDefs "in" eXpr        { DDefVertInit $1 $2 $3 "" }
  / fUn "v" "=" eXpr                            { DDefVertInit $1 [] $2 "" }

dEfGraphVar :: DDefGraphVar String
  = vAr "=" gRaphExpr                           { DDefGraphVar $1 $2 "" }

dEfGraphFun :: DDefGraphFun String
  = fUn vAr "=" "let" dEfGraphVars "in" gRaphExpr  { DDefGraphFun $1 $2 $3 $4 "" }
  / fUn vAr "=" gRaphExpr                      { DDefGraphFun $1 $2 [] $3 "" }
    
dEfGraphVars :: [DDefGraphVar String]
  = dEfGraphVar (";" dEfGraphVar)*              { $1:$2 }
 
gRaphExpr :: DGraphExpr String
  = "fregel" fUn fUn tErmination vAr            { DPregel $1 $2 $3 (DGVar $4 "") "" }
  / "gmap" fUn vAr                              { DGMap $1 (DGVar $2 "") "" }
  / "gzip" vAr vAr                              { DGZip (DGVar $1 "") (DGVar $2 "") "" }
  / "giter" fUn fUn tErmination vAr             { DGIter $1 $2 $3 (DGVar $4 "") "" }
  / vAr                                         { DGVar $1 "" }

tErmination :: DTermination String
  = "Fix"                                       { DTermF ""    }
  / "(" "Iter" eXpr ")"                         { DTermI $1 "" }
  / "(" "Until" pRedExpr ")"                    { DTermU $1 "" }
  / "(" "While" pRedExpr ")"                    { DTermU (DFunAp (DFun "not" "") [$1] "") "" } 
  / "(" tErmination ")"                         { $1       }
 
-- TODO: add UntilAny, UntilAll

pRedExpr :: DExpr String
= "(" "\\" "g" "->" eXpr ")"                    { $1 }

pRedVExpr :: DExpr String
= "(" "\\" "v" "->" eXpr ")"                    { $1 }

sMplDef :: DSmplDef String
  = dEfFun                                      { $1 }
  / dEfVar                                      { $1 }
  / dEfTuple                                    { $1 }
 
dEfFun :: DSmplDef String
  = fUn vAr+ "=" "let" sMplDefs "in" eXpr       { DDefFun $1 $2 $3 $4 "" }
  / fUn vAr+ "=" eXpr                           { DDefFun $1 $2 [] $3 "" }

dEfVar :: DSmplDef String
  = vAr "=" "let" sMplDefs "in" eXpr            { DDefVar $1 $2 $3 "" }
  / vAr "=" eXpr                                { DDefVar $1 [] $2 "" }

dEfTuple :: DSmplDef String
  = "(" vAr ("," vAr)+ ")" "=" "let" sMplDefs "in" eXpr { DDefTuple ($1:$2) $3 $4 "" }
  / "(" vAr ("," vAr)+ ")" "=" eXpr             { DDefTuple ($1:$2) [] $3 "" }

sMplDefs :: [DSmplDef String]
  = sMplDef (";" sMplDef)*                      { $1:$2 } 

-- expression hierarchy

eXpr :: DExpr String
 = eXpr8                                  { $1 }

eXpr8 :: DExpr String
  = "if" eXpr7 "then" eXpr7 "else" eXpr7 { DIf $1 $2 $3 "" }    
  / eXpr7                                { $1 }

eXpr7 :: DExpr String
  = eXpr6 "||" eXpr7                     { DFunAp (DBinOp "||" "") [$1,$2] "" }
  / eXpr6                                { $1 }

eXpr6 :: DExpr String
  = eXpr5 "&&" eXpr6                     { DFunAp (DBinOp "&&" "") [$1,$2] "" }
  / eXpr5                                { $1 }

eXpr5 :: DExpr String
  = eXpr4 "<=" eXpr4                     { DFunAp (DBinOp "<=" "") [$1,$2] "" }
  / eXpr4 ">=" eXpr4                     { DFunAp (DBinOp ">=" "") [$1,$2] "" }
  / eXpr4 "==" eXpr4                     { DFunAp (DBinOp "==" "") [$1,$2] "" }
  / eXpr4 "/=" eXpr4                     { DFunAp (DBinOp "!=" "") [$1,$2] "" }
  / eXpr4 "!=" eXpr4                     { DFunAp (DBinOp "!=" "") [$1,$2] "" }
  / eXpr4 ">" eXpr4                      { DFunAp (DBinOp ">" "") [$1,$2] "" }
  / eXpr4 "<" eXpr4                      { DFunAp (DBinOp "<" "") [$1,$2] "" }
  / eXpr4                                { $1 }

eXpr4 :: DExpr String
  = eXpr3 "`" fUn "`" eXpr3              { DFunAp $2 [$1,$3] "" }
  / eXpr3 (op4 eXpr3)*                   { leftAssoc $1 $2 }
--  / eXpr3                                { $1 }

op4 :: String
  = '+'                                  { "+" }
  / '-'                                  { "-" }

eXpr3 :: DExpr String
  = eXpr2 (op3 eXpr2)*                   { leftAssoc $1 $2 }
  / !cOnstNeg "-" eXpr2                     { DFunAp (DFun "neg" "") [$1] "" }
--  / eXpr2                                { $1 }

op3 :: String
  = '*'                                  { "*" }
  / '/'                                  { "/" }

eXpr2 :: DExpr String
  = tAbleExpr (".^" fIeld)*              { DFieldAcc $1 $2 "" }
  / eDge (".^" fIeld)*                   { DFieldAccE $1 $2 "" }
  / eXpr1                                { $1 }
{-
  = tAbleExpr (".^" fStsNd)* ".^" fIeld { DFieldAcc $1 ($2++[$3]) "" }
  / tAbleExpr                           { DFieldAcc $1 [] "" }
  / eDge (".^" fStsNd)* ".^" fIeld      { DFieldAccE $1 ($2++[$3]) "" }
  / eXpr1                                { $1 }

fStsNd :: DField String
  = "_fst"                               { DFfst "" } 
  / "_snd"                               { DFsnd "" }

 -}

eDge :: DEdge String
 = "e"                                   { DEdge "" }

tAbleExpr :: DTableExpr String
  = "curr" vAr                           { DCurr $1 "" }
  / "prev" vAr                           { DPrev $1 "" }
  / "val"  vAr                           { DVal  $1 "" }
 
eXpr1 :: DExpr String
  = fUn eXpr0+                           { DFunAp $1 $2 "" }
  / cOnstructor eXpr0+                   { DConsAp $1 $2 "" }
  / aGgD "[" eXpr "|" gEn ("," eXpr)* "]" { DAggr $1 $2 $3 $4 "" }
  / aGg "[" eXpr "|" gEn ("," eXpr)* "]" { DAggr $1 $2 $3 $4 "" }
  / eXpr0                                { $1 }

eXpr0 :: DExpr String 
  = cOnst                                { DCExp $1 ""       }
  / cOnstNeg                             { DCExp $1 ""       }
  / vAr                                  { DVExp $1 ""       }
  / "(" eXpr ")"                         { $1               }
  / "(" eXpr ("," eXpr )+ ")"            { DTuple ($1:$2) "" }


gEn :: DGen String
  = "(" "e" "," "u" ")" "<-" "is" "v" { DGenI "" }
  / "(" "e" "," "u" ")" "<-" "rs" "v" { DGenO "" }
  / "u" "<-" "gof" "v"                { DGenG "" }
  / "u" "<-" "g"                      { DGenG "" }

aGg ::: DAgg String
   = "minimum" { DAggMin ""  }
   / "maximum" { DAggMax ""  }
   / "sum"     { DAggSum ""  }
   / "prod"    { DAggProd "" }
   / "and"     { DAggAnd ""  }
   / "or"      { DAggOr ""   }

aGgD ::: DAgg String
   = "random" eXpr { DAggChoice $1 ""  }

cOnstructor ::: DConstructor String
   = !reserved [A-Z][a-zA-Z_0-9']* { DConstructor ($1:$2) "" }

vAr ::: DVar String
   = !reserved [a-z_][a-zA-Z_0-9']* { DVar ($1:$2) "" }

fIeld ::: DField String
   = !reserved [a-z_][a-zA-Z_0-9']* { DField ($1:$2) "" }

fUn ::: DFun String
   = !reserved [a-z_][a-zA-Z_0-9']* { DFun ($1:$2) "" }

reserved :: ()
  = "if"
  / "then"
  / "else"
  / "let"
  / "in"
  / "val"
  / "curr"
  / "prev"
  / "minimum"
  / "maximum"
  / "random"
  / "sum"
  / "prod"
  / "and"
  / "or"
  / "is"
  / "rs"
  / "fregel"
  / "gmap"
  / "gzip"
  / "giter"
  / "True"
  / "False"
  / "Int"
  / "Bool"
  / "Pair"
  / "Double"
                         
cOnst ::: DConst String
   = [1-9][0-9]* '.' [0-9]+ expo { DCDouble (read ($1 : $2++"."++$3++$4)) "" }
   / '0' '.' [0-9]+ 'e' expo     { DCDouble (read ("0."++$1++$2)) ""         }
   / [1-9][0-9]* '.' [0-9]+      { DCDouble (read ($1 : $2++"."++$3)) ""     }
   / '0' '.' [0-9]+              { DCDouble (read ("0."++$1)) ""             }
--   / "-" '0' '.' [0-9]+         { DCDouble (- (read ("0."++$1))) ""         }
--   / "-" [1-9][0-9]* '.' [0-9]+ { DCDouble (- (read ($1 : $2++"."++$3))) "" }
--   / "-" "0.0"                  { DCDouble (-0.0) ""                        }   
--   / "-" [1-9][0-9]*            { DCInt (- (read  ($1 : $2))) ""            }
   / [1-9][0-9]*                { DCInt (read  ($1 : $2)) ""                }
--   / "-" "0"                    { DCInt 0 ""                                }
   / "0"                        { DCInt 0 ""                                }
   / "True"                     { DCBool True ""                            }
   / "False"                    { DCBool False ""                           }
   / '\"' [^\"]* '\"'           { DCString $1 ""                            }

cOnstNeg ::: DConst String
   = "-" '0' '.' [0-9]+         { DCDouble (- (read ("0."++$1))) ""         }
   / "-" [1-9][0-9]* '.' [0-9]+ { DCDouble (- (read ($1 : $2++"."++$3))) "" }
   / "-" "0.0"                  { DCDouble (-0.0) ""                        }   
   / "-" [1-9][0-9]*            { DCInt (- (read  ($1 : $2))) ""            }
   / "-" "0"                    { DCInt 0 ""                                }

expo :: String
 = 'e' [0-9]+       { "e" ++ $1 }
 / 'E' [0-9]+       { "e" ++ $1 }
 / 'e' '-' [0-9]+   { "e-" ++ $1 }
 / 'E' '-' [0-9]+   { "e-" ++ $1 }

           
tYpe ::: DType String
   = "Int"                    { DTInt "" }
   / "Bool"                   { DTBool "" }
   / "String"                 { DTString "" }
   / "Double"                 { DTDouble "" }
   / "Pair" tYpe tYpe         { DTRecord (DConstructor "Pair" "") [$1, $2] "" }
   / "(" tYpe ("," tYpe)+ ")" { DTTuple ($1:$2) "" }
   / "(" tYpe  ")"            { $1 }
   / cOnstructor              { DTRecord $1 [] "" }

comment :: ()
  = '{-' (comment / (!"-}" . { () } ))* "-}" { () }
  
space :: () = [ \r\n\t] { () } / comment { () }

|]


{- 
checking function: parse = parse . prettyPrint . parse ?

-}

check file = 
  do ast <- parseFile pRogramSpec file
     let pp = ppAST (right ast)
     let ast2 = parseString pRogramSpec "" pp
     return (right ast == right ast2)

--- misc functions to be used by other modules

processFile file =
  do ast <- parseFile pRogramSpec file
     print ast

parseFile' file = 
  do ast <- parseFile pRogramSpec file
     return (right ast)

parseFile'' file = 
  do asts <- parseFile pRogramSpecs file
     return (propRecs (right asts))
  where propRecs xs =
          let rss = concatMap (\(DProgramSpec rs _ _) -> rs) xs
          in map (\(DProgramSpec _ x a) -> DProgramSpec rss x a) xs

parseString' ss = right $ parseString pRogramSpec "" ss

parseStringExpr ss = right $ parseString eXpr "" ss
